@@ -33,6 +33,7 @@
 #include "core-format.h"
 #include "internal.h"
 
+#define MIN_SAMPLES     24u
 #define MIN_BUFFERS     8u
 #define MAX_BUFFERS     64u
 
@@ -51,39 +52,6 @@ static void dump_buffer_attr(pa_stream *s, pa_buffer_attr *attr)
 	pw_log_debug("stream %p: fragsize: %u", s, attr->fragsize);
 }
 
-static void configure_device(pa_stream *s)
-{
-	struct global *g;
-	const char *str;
-	uint32_t old = s->device_index;
-
-	g = pa_context_find_linked(s->context, pa_stream_get_index(s));
-	if (g == NULL) {
-		s->device_index = PA_INVALID_INDEX;
-		s->device_name = NULL;
-	} else {
-		if (s->direction == PA_STREAM_RECORD) {
-			if (g->mask == (PA_SUBSCRIPTION_MASK_SINK | PA_SUBSCRIPTION_MASK_SOURCE))
-				s->device_index = g->node_info.monitor;
-			else
-				s->device_index = g->id;
-		}
-		else {
-			s->device_index = g->id;
-		}
-
-		free(s->device_name);
-		if ((str = pw_properties_get(g->props, PW_KEY_NODE_NAME)) == NULL)
-			s->device_name = strdup("unknown");
-		else
-			s->device_name = strdup(str);
-	}
-	pw_log_debug("stream %p: linked to %d '%s'", s, s->device_index, s->device_name);
-
-	if (old != s->device_index && s->moved_callback)
-		s->moved_callback(s, s->moved_userdata);
-}
-
 static void stream_destroy(void *data)
 {
 	pa_stream *s = data;
@@ -119,16 +87,14 @@ static void stream_state_changed(void *data, enum pw_stream_state old,
 		s->stream_index = pw_stream_get_node_id(s->stream);
 		if (!s->suspended) {
 			s->suspended = true;
-			if (!c->disconnect && s->state == PA_STREAM_READY && s->suspended_callback)
+			if (!c->disconnect && !s->corked && s->state == PA_STREAM_READY && s->suspended_callback)
 				s->suspended_callback(s, s->suspended_userdata);
 		}
 		break;
 	case PW_STREAM_STATE_STREAMING:
-		configure_device(s);
-		pa_stream_set_state(s, PA_STREAM_READY);
 		if (s->suspended) {
 			s->suspended = false;
-			if (!c->disconnect && s->started_callback)
+			if (!c->disconnect && !s->corked && s->state == PA_STREAM_READY && s->started_callback)
 				s->started_callback(s, s->started_userdata);
 		}
 		break;
@@ -147,7 +113,7 @@ static const struct spa_pod *get_buffers_param(pa_stream *s, pa_buffer_attr *att
 	size = attr->minreq;
 	buffers = SPA_CLAMP(maxsize / size, MIN_BUFFERS, MAX_BUFFERS);
 
-	pw_log_debug("stream %p: stride %d maxsize %d size %u buffers %d", s, stride, maxsize,
+	pw_log_info("stream %p: stride %d maxsize %d size %u buffers %d", s, stride, maxsize,
 			size, buffers);
 
 	param = spa_pod_builder_add_object(b,
@@ -164,7 +130,7 @@ static const struct spa_pod *get_buffers_param(pa_stream *s, pa_buffer_attr *att
 static void patch_buffer_attr(pa_stream *s, pa_buffer_attr *attr, pa_stream_flags_t *flags) {
 	const char *e, *str;
 	char buf[100];
-	uint32_t stride;
+	uint32_t stride, period;
 
 	pa_assert(s);
 	pa_assert(attr);
@@ -209,31 +175,57 @@ static void patch_buffer_attr(pa_stream *s, pa_buffer_attr *attr, pa_stream_flag
 				*flags |= PA_STREAM_ADJUST_LATENCY;
 		}
 	}
+
+	if (flags && !SPA_FLAG_IS_SET(*flags, PA_STREAM_ADJUST_LATENCY)) {
+		if (attr->maxlength == 0)
+			attr->maxlength = -1;
+		if (attr->tlength == 0)
+			attr->tlength = -1;
+		if (attr->minreq == 0)
+			attr->minreq = -1;
+		if (attr->prebuf == 0)
+			attr->prebuf = -1;
+		if (attr->fragsize == 0)
+			attr->fragsize = -1;
+		period = 100;
+	} else {
+		period = 20;
+	}
+
 	dump_buffer_attr(s, attr);
 
 	stride  = pa_frame_size(&s->sample_spec);
 	if (attr->maxlength == (uint32_t) -1 || attr->maxlength == 0)
 		attr->maxlength = MAX_SIZE; /* 4MB is the maximum queue length PulseAudio <= 0.9.9 supported. */
+	attr->maxlength -= attr->maxlength % stride;
+	attr->maxlength = SPA_MAX(attr->maxlength, stride);
 
-	if (attr->tlength == (uint32_t) -1 || attr->tlength == 0)
+	if (attr->tlength == (uint32_t) -1)
 		attr->tlength = (uint32_t) pa_usec_to_bytes(2*PA_USEC_PER_SEC, &s->sample_spec);
 	attr->tlength = SPA_MIN(attr->tlength, attr->maxlength);
+	attr->tlength -= attr->tlength % stride;
 
-	if (attr->minreq == (uint32_t) -1 || attr->minreq == 0)
-		attr->minreq = pa_usec_to_bytes(25*PA_USEC_PER_MSEC, &s->sample_spec);
-	attr->minreq = SPA_MIN(attr->minreq, attr->tlength / MIN_BUFFERS);
+	if (attr->minreq == (uint32_t) -1)
+		attr->minreq = pa_usec_to_bytes(period*PA_USEC_PER_MSEC, &s->sample_spec);
+	attr->minreq = SPA_MIN(attr->minreq, attr->tlength / 4);
+	attr->minreq = SPA_MAX(attr->minreq, MIN_SAMPLES * stride);
+	attr->minreq -= attr->minreq % stride;
 	attr->minreq = SPA_MAX(attr->minreq, stride);
 
-	if (attr->fragsize == (uint32_t) -1 || attr->fragsize == 0)
-		attr->fragsize = pa_usec_to_bytes(25*PA_USEC_PER_MSEC, &s->sample_spec);
-	attr->fragsize = SPA_MIN(attr->fragsize, attr->tlength / MIN_BUFFERS);
-	attr->fragsize = SPA_MAX(attr->fragsize, stride);
+	attr->tlength = SPA_MAX(attr->tlength, attr->minreq * 4);
 
-	if (attr->prebuf == (uint32_t) -1 || attr->prebuf == 0)
+	if (attr->prebuf == (uint32_t) -1)
 		attr->prebuf = attr->tlength - attr->minreq;
 	attr->prebuf = SPA_MIN(attr->prebuf, attr->tlength - attr->minreq);
+	attr->prebuf -= attr->prebuf % stride;
 	attr->prebuf = SPA_MAX(attr->prebuf, stride);
 
+	if (attr->fragsize == (uint32_t) -1)
+		attr->fragsize = pa_usec_to_bytes(period*PA_USEC_PER_MSEC, &s->sample_spec);
+	attr->fragsize = SPA_MIN(attr->fragsize, attr->tlength / 4);
+	attr->fragsize -= attr->fragsize % stride;
+	attr->fragsize = SPA_MAX(attr->fragsize, stride);
+
 	dump_buffer_attr(s, attr);
 }
 
@@ -241,9 +233,11 @@ static void stream_param_changed(void *data, uint32_t id, const struct spa_pod *
 {
 	pa_stream *s = data;
 	const struct spa_pod *params[4];
-	uint32_t n_params = 0;
-        uint8_t buffer[4096];
-        struct spa_pod_builder b = SPA_POD_BUILDER_INIT(buffer, sizeof(buffer));
+	uint32_t n_params = 0, stride, latency;
+	uint8_t buffer[4096];
+	struct spa_pod_builder b = SPA_POD_BUILDER_INIT(buffer, sizeof(buffer));
+	struct spa_dict_item items[1];
+	char str[64];
 	int res;
 
 	if (param == NULL || id != SPA_PARAM_Format)
@@ -257,11 +251,30 @@ static void stream_param_changed(void *data, uint32_t id, const struct spa_pod *
 	if (s->format)
 		pa_format_info_free(s->format);
 	s->format = pa_format_info_from_sample_spec(&s->sample_spec, &s->channel_map);
+	if (s->format == NULL) {
+		pw_stream_set_error(s->stream, -errno, "unhandled format");
+		return;
+	}
 
-	patch_buffer_attr(s, &s->buffer_attr, NULL);
+	patch_buffer_attr(s, &s->buffer_attr, &s->flags);
+
+	pa_stream_set_state(s, PA_STREAM_READY);
+
+	if (s->corked)
+		pw_stream_set_active(s->stream, false);
+
+	stride = pa_frame_size(&s->sample_spec);
+
+	if (s->direction == PA_STREAM_RECORD) {
+		latency = s->buffer_attr.fragsize / stride;
+	} else {
+		latency = s->buffer_attr.minreq * 2 / stride;
+	}
+	snprintf(str, sizeof(str), "%u/%u", latency, s->sample_spec.rate);
+	items[0] = SPA_DICT_ITEM_INIT(PW_KEY_NODE_LATENCY, str);
+	pw_stream_update_properties(s->stream, &SPA_DICT_INIT(items, 1));
 
 	params[n_params++] = get_buffers_param(s, &s->buffer_attr, &b);
-
 	pw_stream_update_params(s->stream, params, n_params);
 }
 
@@ -287,9 +300,10 @@ static void stream_control_info(void *data, uint32_t id, const struct pw_stream_
 static void stream_add_buffer(void *data, struct pw_buffer *buffer)
 {
 	pa_stream *s = data;
-	buffer->size = buffer->buffer->datas[0].maxsize;
-	s->maxsize += buffer->size;
-	s->maxblock = SPA_MIN(buffer->size, s->maxblock);
+	uint32_t maxsize = buffer->buffer->datas[0].maxsize;
+	buffer->size = 0;
+	s->maxsize += maxsize;
+	s->maxblock = SPA_MIN(maxsize, s->maxblock);
 }
 
 static void stream_remove_buffer(void *data, struct pw_buffer *buffer)
@@ -324,10 +338,11 @@ static void update_timing_info(pa_stream *s)
 	ti->read_index_corrupt = false;
 
 	if (pwt.rate.denom > 0) {
+		uint64_t ticks = pwt.ticks;
 		if (!s->have_time)
-			s->ticks_base = pwt.ticks + pwt.delay;
-		if (pwt.ticks > s->ticks_base)
-			pos = ((pwt.ticks - s->ticks_base) * s->sample_spec.rate / pwt.rate.denom) * stride;
+			s->ticks_base = ticks;
+		if (ticks > s->ticks_base)
+			pos = ((ticks - s->ticks_base) * s->sample_spec.rate / pwt.rate.denom) * stride;
 		else
 			pos = 0;
 		delay = pwt.delay * SPA_USEC_PER_SEC / pwt.rate.denom;
@@ -345,12 +360,12 @@ static void update_timing_info(pa_stream *s)
 		ti->configured_source_usec = delay;
 		ti->write_index = pos;
 	}
-	s->timing_info_valid = true;
 	s->queued_bytes = pwt.queued;
+	s->timing_info_valid = true;
 
 	pw_log_trace("stream %p: %"PRIu64" rate:%d/%d ticks:%"PRIu64" pos:%"PRIu64" delay:%"PRIi64 " read:%"PRIu64
 			" write:%"PRIu64" queued:%"PRIi64,
-			s, pwt.queued, s->sample_spec.rate, pwt.rate.denom, pwt.ticks, pos, delay,
+			s, pwt.queued, s->sample_spec.rate, pwt.rate.denom, pwt.ticks, pos, pwt.delay,
 			ti->read_index, ti->write_index, ti->read_index - ti->write_index);
 }
 
@@ -373,12 +388,14 @@ static void queue_output(pa_stream *s)
 		pw_log_trace("queue %p", m);
 		spa_list_remove(&m->link);
 		s->ready_bytes -= m->size;
+		s->queued_bytes += m->size;
 
 		buf->buffer->datas[0].maxsize = m->maxsize;
 		buf->buffer->datas[0].data = m->data;
 		buf->buffer->datas[0].chunk->offset = m->offset;
 		buf->buffer->datas[0].chunk->size = m->size;
 		buf->user_data = m;
+		buf->size = m->size;
 		m->user_data = buf;
 
 		pw_stream_queue_buffer(s->stream, buf);
@@ -423,12 +440,56 @@ static void pull_input(pa_stream *s)
 		m->user_data = buf;
 		buf->user_data = m;
 
-		pw_log_trace("input %p", m);
+		pw_log_trace("input %p, size:%zd ready:%zd", m, m->size, s->ready_bytes);
 		spa_list_append(&s->ready, &m->link);
 		s->ready_bytes += m->size;
 	}
 }
 
+
+static inline uint32_t queued_size(const pa_stream *s, uint64_t elapsed)
+{
+	uint64_t queued;
+	const pa_timing_info *i = &s->timing_info;
+	queued = i->write_index - SPA_MIN(i->read_index, i->write_index);
+	queued -= SPA_MIN(queued, elapsed);
+	return queued;
+}
+static inline uint32_t target_queue(const pa_stream *s)
+{
+	return s->buffer_attr.tlength;
+}
+
+static inline uint32_t wanted_size(const pa_stream *s, uint32_t queued, uint32_t target)
+{
+	return target - SPA_MIN(queued, target);
+}
+
+static inline uint32_t required_size(const pa_stream *s)
+{
+	return s->buffer_attr.minreq;
+}
+
+static inline uint32_t writable_size(const pa_stream *s, uint64_t elapsed)
+{
+	uint32_t queued, target, wanted, required;
+
+	queued = queued_size(s, elapsed);
+	target = target_queue(s);
+	wanted = wanted_size(s, queued, target);
+	required = required_size(s);
+
+	pw_log_trace("stream %p, queued:%u target:%u wanted:%u required:%u",
+			s, queued, target, wanted, required);
+	if (SPA_FLAG_IS_SET(s->flags, PA_STREAM_ADJUST_LATENCY))
+		if (queued >= wanted)
+			wanted = 0;
+	if (wanted < required)
+		wanted = 0;
+
+	return wanted;
+}
+
 static void stream_process(void *data)
 {
 	pa_stream *s = data;
@@ -437,16 +498,13 @@ static void stream_process(void *data)
 	update_timing_info(s);
 
 	if (s->direction == PA_STREAM_PLAYBACK) {
-		pa_timing_info *i = &s->timing_info;
-		uint64_t queued, writable, required;
+		uint32_t writable;
 
 		queue_output(s);
 
-		queued = i->write_index - SPA_MIN(i->read_index, i->write_index);
-		writable = s->maxblock - SPA_MIN(queued, s->maxblock);
-		required = SPA_MIN(s->maxblock, s->buffer_attr.minreq);
+		writable = writable_size(s, 0);
 
-		if (s->write_callback && s->state == PA_STREAM_READY && writable >= required)
+		if (s->write_callback && s->state == PA_STREAM_READY && writable > 0)
 			s->write_callback(s, writable, s->write_userdata);
 	}
 	else {
@@ -533,13 +591,17 @@ static pa_stream* stream_new(pa_context *c, const char *name,
 	else
 		pa_channel_map_init(&s->channel_map);
 
-	pw_log_debug("channel map: %p %s", map, pa_channel_map_snprint(str, sizeof(str), &s->channel_map));
+	pw_log_debug("stream %p: channel map: %p %s", s,
+			map, pa_channel_map_snprint(str, sizeof(str), &s->channel_map));
 
 	s->n_formats = 0;
 	if (formats) {
 		s->n_formats = n_formats;
-		for (i = 0; i < n_formats; i++)
+		for (i = 0; i < n_formats; i++) {
 			s->req_formats[i] = pa_format_info_copy(formats[i]);
+			pw_log_debug("format %d: %s", i,
+					pa_format_info_snprint(str, sizeof(str), formats[i]));
+		}
 	}
 	s->format = NULL;
 
@@ -598,12 +660,12 @@ static void stream_unlink(pa_stream *s)
 		if (o->stream == s)
 			pa_operation_cancel(o);
 	}
+	s->drain = NULL;
 
 	spa_list_remove(&s->link);
 	if (s->stream)
 		pw_stream_set_active(s->stream, false);
 
-	s->stream_index = PA_INVALID_INDEX;
 	s->context = NULL;
 	pa_stream_unref(s);
 }
@@ -615,10 +677,13 @@ static void stream_free(pa_stream *s)
 
 	pw_log_debug("stream %p", s);
 
+
 	if (s->stream) {
 		spa_hook_remove(&s->stream_listener);
 		pw_stream_destroy(s->stream);
 	}
+	if (s->global)
+		s->global->stream = NULL;
 
 	spa_list_consume(m, &s->free, link) {
 		pw_log_trace("free %p", m);
@@ -751,7 +816,7 @@ int pa_stream_is_suspended(PA_CONST pa_stream *s)
 	PA_CHECK_VALIDITY(s->context, s->state == PA_STREAM_READY, PA_ERR_BADSTATE);
 	PA_CHECK_VALIDITY(s->context, s->direction != PA_STREAM_UPLOAD, PA_ERR_BADSTATE);
 
-	return s->suspended;
+	return s->suspended && !s->corked;
 }
 
 SPA_EXPORT
@@ -777,15 +842,13 @@ static int create_stream(pa_stream_direction_t direction,
 {
 	int res;
 	enum pw_stream_flags fl;
-	const struct spa_pod *params[16];
-	uint32_t i, n_params = 0, stride;
+	const struct spa_pod *params[PA_MAX_FORMATS+1];
+	uint32_t i, n_params = 0;
 	uint8_t buffer[4096];
 	struct spa_pod_builder b = SPA_POD_BUILDER_INIT(buffer, sizeof(buffer));
 	const char *str;
 	uint32_t devid, n_items;
-	struct global *g;
 	struct spa_dict_item items[7];
-	char latency[64];
 	bool monitor, no_remix;
 	const char *name;
 	pa_context *c = s->context;
@@ -836,7 +899,7 @@ static int create_stream(pa_stream_direction_t direction,
 	s->disconnecting = false;
 	if (volume) {
 		for (i = 0; i < volume->channels; i++)
-			s->channel_volumes[i] = volume->values[i] / (float) PA_VOLUME_NORM;
+			s->channel_volumes[i] = pa_sw_volume_to_linear(volume->values[i]);
 		s->n_channel_volumes = volume->channels;
 	} else {
 		for (i = 0; i < SPA_AUDIO_MAX_CHANNELS; i++)
@@ -844,6 +907,7 @@ static int create_stream(pa_stream_direction_t direction,
 		s->n_channel_volumes = 0;
 	}
 	s->mute = false;
+	s->flags = flags;
 
 	pa_stream_set_state(s, PA_STREAM_CREATING);
 
@@ -852,8 +916,6 @@ static int create_stream(pa_stream_direction_t direction,
 
 	s->corked = SPA_FLAG_IS_SET(flags, PA_STREAM_START_CORKED);
 
-	if (s->corked)
-		fl |= PW_STREAM_FLAG_INACTIVE;
 	if (flags & PA_STREAM_PASSTHROUGH)
 		fl |= PW_STREAM_FLAG_EXCLUSIVE;
 	if (flags & PA_STREAM_DONT_MOVE)
@@ -872,13 +934,12 @@ static int create_stream(pa_stream_direction_t direction,
 		pa_sample_spec ss;
 		pa_channel_map chmap;
 		int i;
-		uint32_t sample_rate = 0;
 
 		for (i = 0; i < s->n_formats; i++) {
 			if ((res = pa_format_info_to_sample_spec(s->req_formats[i], &ss, NULL)) < 0) {
 				char buf[4096];
 				pw_log_warn("can't convert format %d %s", res,
-						pa_format_info_snprint(buf,4096,s->req_formats[i]));
+						pa_format_info_snprint(buf, sizeof(buf), s->req_formats[i]));
 				continue;
 			}
 			if (pa_format_info_get_channel_map(s->req_formats[i], &chmap) < 0)
@@ -886,79 +947,59 @@ static int create_stream(pa_stream_direction_t direction,
 
 			params[n_params++] = pa_format_build_param(&b, SPA_PARAM_EnumFormat,
 					&ss, &chmap);
-			if (ss.rate > sample_rate) {
-				sample_rate = ss.rate;
-				s->sample_spec = ss;
-			}
-		}
-		if (sample_rate == 0) {
-			s->sample_spec.format = PA_SAMPLE_S16NE;
-			s->sample_spec.rate = 48000;
-			s->sample_spec.channels = 2;
 		}
 	}
-	if (!pa_sample_spec_valid(&s->sample_spec))
-		return -EINVAL;
-
-	patch_buffer_attr(s, &s->buffer_attr, &flags);
 
 	if (direction == PA_STREAM_RECORD)
 		devid = s->direct_on_input;
 	else
 		devid = PW_ID_ANY;
 
-	if (dev == NULL) {
-		if ((str = getenv("PIPEWIRE_NODE")) != NULL)
-			devid = atoi(str);
+	if (dev == NULL && devid == PW_ID_ANY) {
+		dev = getenv("PIPEWIRE_NODE");
 	}
-	else if (devid == PW_ID_ANY) {
-		uint32_t mask;
-
-		if (direction == PA_STREAM_PLAYBACK)
-			mask = PA_SUBSCRIPTION_MASK_SINK;
-		else if (direction == PA_STREAM_RECORD)
-			mask = PA_SUBSCRIPTION_MASK_SOURCE;
-		else
-			mask = 0;
-
-		if ((g = pa_context_find_global_by_name(s->context, mask, dev)) != NULL)
-			devid = g->id;
-		else if ((devid = atoi(dev)) == 0)
+	else if (dev != NULL && devid == PW_ID_ANY) {
+		if ((devid = atoi(dev)) == 0)
 			devid = PW_ID_ANY;
+		else if (devid & PA_IDX_FLAG_MONITOR)
+			devid &= PA_IDX_MASK_MONITOR;
+
+		if (devid == PW_ID_ANY) {
+			if (pa_endswith(dev, ".monitor"))
+				dev = strndupa(dev, strlen(dev) - 8);
+		}
 	}
 
-	if ((str = pa_proplist_gets(s->proplist, PA_PROP_MEDIA_ROLE)) == NULL)
-		str = "Music";
-	else if (strcmp(str, "video") == 0)
-		str = "Movie";
-	else if (strcmp(str, "music") == 0)
-		str = "Music";
-	else if (strcmp(str, "game") == 0)
-		str = "Game";
-	else if (strcmp(str, "event") == 0)
-		str = "Notification";
-	else if (strcmp(str, "phone") == 0)
-		str = "Communication";
-	else if (strcmp(str, "animation") == 0)
-		str = "Movie";
-	else if (strcmp(str, "production") == 0)
-		str = "Production";
-	else if (strcmp(str, "a11y") == 0)
-		str = "Accessibility";
-	else if (strcmp(str, "test") == 0)
-		str = "Test";
-	else
-		str = "Music";
+	if ((str = pa_proplist_gets(s->proplist, PA_PROP_MEDIA_ROLE)) != NULL) {
+		if (strcmp(str, "video") == 0)
+			str = "Movie";
+		else if (strcmp(str, "music") == 0)
+			str = "Music";
+		else if (strcmp(str, "game") == 0)
+			str = "Game";
+		else if (strcmp(str, "event") == 0)
+			str = "Notification";
+		else if (strcmp(str, "phone") == 0)
+			str = "Communication";
+		else if (strcmp(str, "animation") == 0)
+			str = "Movie";
+		else if (strcmp(str, "production") == 0)
+			str = "Production";
+		else if (strcmp(str, "a11y") == 0)
+			str = "Accessibility";
+		else if (strcmp(str, "test") == 0)
+			str = "Test";
+		else
+			str = "Music";
+	}
 
-	stride = pa_frame_size(&s->sample_spec);
-	sprintf(latency, "%u/%u", s->buffer_attr.minreq / stride, s->sample_spec.rate);
 	n_items = 0;
-	items[n_items++] = SPA_DICT_ITEM_INIT(PW_KEY_NODE_LATENCY, latency);
 	items[n_items++] = SPA_DICT_ITEM_INIT(PW_KEY_MEDIA_TYPE, "Audio");
 	items[n_items++] = SPA_DICT_ITEM_INIT(PW_KEY_MEDIA_CATEGORY,
 				direction == PA_STREAM_PLAYBACK ?
 					"Playback" : monitor ? "Monitor" : "Capture");
-	items[n_items++] = SPA_DICT_ITEM_INIT(PW_KEY_MEDIA_ROLE, str);
+	if (str != NULL)
+		items[n_items++] = SPA_DICT_ITEM_INIT(PW_KEY_MEDIA_ROLE, str);
 	if (monitor)
 		items[n_items++] = SPA_DICT_ITEM_INIT(PW_KEY_STREAM_MONITOR, "true");
 	if (no_remix)
@@ -1024,6 +1065,7 @@ int pa_stream_disconnect(pa_stream *s)
 	pa_stream_ref(s);
 
 	s->disconnecting = true;
+	s->stream_index = PA_INVALID_INDEX;
 	pw_stream_disconnect(s->stream);
 
 	o = pa_operation_new(c, s, on_disconnected, 0);
@@ -1232,7 +1274,7 @@ SPA_EXPORT
 size_t pa_stream_writable_size(PA_CONST pa_stream *s)
 {
 	const pa_timing_info *i;
-	uint64_t now, then, queued, writable, elapsed, required;
+	uint64_t now, then, elapsed;
 	struct timespec ts;
 
 	spa_assert(s);
@@ -1254,18 +1296,7 @@ size_t pa_stream_writable_size(PA_CONST pa_stream *s)
 		elapsed = 0;
 	}
 
-	queued = i->write_index - SPA_MIN(i->read_index, i->write_index);
-	queued -= SPA_MIN(queued, elapsed);
-
-	writable = s->maxblock - SPA_MIN(queued, s->maxblock);
-	required = SPA_MIN(s->maxblock, s->buffer_attr.minreq);
-
-	pw_log_debug("stream %p: %"PRIu64" minreq:%u maxblock:%zu", s,
-			writable, s->buffer_attr.minreq, s->maxblock);
-	if (writable < required)
-		writable = 0;
-
-	return writable;
+	return writable_size(s, elapsed);
 }
 
 SPA_EXPORT
@@ -1314,6 +1345,10 @@ pa_operation* pa_stream_drain(pa_stream *s, pa_stream_success_cb_t cb, void *use
 	PA_CHECK_VALIDITY_RETURN_NULL(s->context, s->direction == PA_STREAM_PLAYBACK, PA_ERR_BADSTATE);
 
 	pw_log_debug("stream %p", s);
+	if (s->corked) {
+		s->corked = false;
+		pw_stream_set_active(s->stream, true);
+	}
 	pw_stream_flush(s->stream, true);
 	o = pa_operation_new(s->context, s, on_success, sizeof(struct success_ack));
 	d = o->userdata;
@@ -1354,6 +1389,7 @@ pa_operation* pa_stream_update_timing_info(pa_stream *s, pa_stream_success_cb_t
 	PA_CHECK_VALIDITY_RETURN_NULL(s->context, s->state == PA_STREAM_READY, PA_ERR_BADSTATE);
 	PA_CHECK_VALIDITY_RETURN_NULL(s->context, s->direction != PA_STREAM_UPLOAD, PA_ERR_BADSTATE);
 
+	pw_log_debug("stream %p", s);
 	o = pa_operation_new(s->context, s, on_timing_success, sizeof(struct success_ack));
 	d = o->userdata;
 	d->cb = cb;
@@ -1567,6 +1603,7 @@ pa_operation* pa_stream_flush(pa_stream *s, pa_stream_success_cb_t cb, void *use
 			b->user_data = NULL;
 	}
 	s->ready_bytes = 0;
+	s->queued_bytes = 0;
 	s->timing_info.write_index = s->timing_info.read_index = 0;
 	s->have_time = false;
 	pa_operation_sync(o);
@@ -1665,10 +1702,7 @@ int pa_stream_get_time(pa_stream *s, pa_usec_t *r_usec)
 
 	if (s->direction == PA_STREAM_PLAYBACK) {
 		res = pa_bytes_to_usec((uint64_t) i->read_index, &s->sample_spec);
-		if (res > i->sink_usec)
-			res -= i->sink_usec;
-		else
-			res = 0;
+		res -= SPA_MIN(res, i->sink_usec);
 	} else {
 		res = pa_bytes_to_usec((uint64_t) i->write_index, &s->sample_spec);
 		res += i->source_usec;
@@ -1684,9 +1718,9 @@ int pa_stream_get_time(pa_stream *s, pa_usec_t *r_usec)
 		*r_usec = res;
 
 	pw_log_trace("stream %p: now:%"PRIu64" diff:%"PRIi64
-			" write-index:%"PRIi64" read_index:%"PRIi64" res:%"PRIu64,
-			s, now, now - res, i->write_index, i->read_index, res);
-
+			" write-index:%"PRIi64" read_index:%"PRIi64" rw-diff:%"PRIi64" res:%"PRIu64,
+			s, now, now - res, i->write_index, i->read_index,
+			i->write_index - i->read_index, res);
 	return 0;
 }
 
@@ -1738,8 +1772,8 @@ int pa_stream_get_latency(pa_stream *s, pa_usec_t *r_usec, int *negative)
 	else
 		*r_usec = time_counter_diff(s, t, c, negative);
 
-	pw_log_trace("stream %p: now:%"PRIu64" stream:%"PRIu64
-			" res:%"PRIu64, s, t, c, *r_usec);
+	pw_log_trace("stream %p: now:%"PRIu64" stream:%"PRIu64" cindex:%"PRIi64
+			" res:%"PRIu64, s, t, c, cindex, *r_usec);
 
 	return 0;
 }
@@ -1899,6 +1933,7 @@ int pa_stream_set_monitor_stream(pa_stream *s, uint32_t sink_input_idx)
 	PA_CHECK_VALIDITY(s->context, sink_input_idx != PA_INVALID_INDEX, PA_ERR_INVALID);
 	PA_CHECK_VALIDITY(s->context, s->state == PA_STREAM_UNCONNECTED, PA_ERR_BADSTATE);
 
+	pw_log_debug("stream %p: Set monitor stream %u", s, sink_input_idx);
 	s->direct_on_input = sink_input_idx;
 	return 0;
 }
@@ -1909,6 +1944,7 @@ uint32_t pa_stream_get_monitor_stream(PA_CONST pa_stream *s)
 	spa_assert(s);
 	spa_assert(s->refcount >= 1);
 
+	pw_log_debug("stream %p: get monitor stream %u", s, s->direct_on_input);
 	PA_CHECK_VALIDITY_RETURN_ANY(s->context, s->direct_on_input != PA_INVALID_INDEX,
 			PA_ERR_BADSTATE, PA_INVALID_INDEX);
 
