@@ -42,14 +42,18 @@ extern "C" {
 #include <spa/node/node.h>
 #include <spa/node/utils.h>
 #include <spa/node/io.h>
+#include <spa/debug/types.h>
 #include <spa/param/param.h>
 #include <spa/param/audio/format-utils.h>
 
+#include "dll.h"
+
 #define MIN_LATENCY	16
 #define MAX_LATENCY	8192
 
 #define DEFAULT_RATE		48000u
 #define DEFAULT_CHANNELS	2u
+#define DEFAULT_USE_CHMAP	false
 
 struct props {
 	char device[64];
@@ -57,6 +61,7 @@ struct props {
 	char card_name[128];
 	uint32_t min_latency;
 	uint32_t max_latency;
+	bool use_chmap;
 };
 
 #define MAX_BUFFERS 32
@@ -101,7 +106,9 @@ struct state {
 	bool have_format;
 	struct spa_audio_info current_format;
 
+	uint32_t default_format;
 	unsigned int default_channels;
+	unsigned int default_rate;
 
 	snd_pcm_uframes_t buffer_frames;
 	snd_pcm_uframes_t period_frames;
@@ -135,6 +142,7 @@ struct state {
 	int timerfd;
 	uint32_t threshold;
 	uint32_t last_threshold;
+	uint32_t headroom;
 
 	uint32_t duration;
 	uint32_t last_duration;
@@ -156,9 +164,7 @@ struct state {
 	uint64_t underrun;
 	double safety;
 
-	double bw;
-	double z1, z2, z3;
-	double w0, w1, w2;
+	struct spa_dll dll;
 };
 
 int
@@ -168,6 +174,7 @@ spa_alsa_enum_format(struct state *state, int seq,
 
 int spa_alsa_set_format(struct state *state, struct spa_audio_info *info, uint32_t flags);
 
+int spa_alsa_open(struct state *state);
 int spa_alsa_start(struct state *state);
 int spa_alsa_reassign_follower(struct state *state);
 int spa_alsa_pause(struct state *state);
@@ -178,6 +185,16 @@ int spa_alsa_read(struct state *state, snd_pcm_uframes_t silence);
 
 void spa_alsa_recycle_buffer(struct state *state, uint32_t buffer_id);
 
+static inline uint32_t spa_alsa_format_from_name(const char *name, size_t len)
+{
+	int i;
+	for (i = 0; spa_type_audio_format[i].name; i++) {
+		if (strncmp(name, spa_debug_type_short_name(spa_type_audio_format[i].name), len) == 0)
+			return spa_type_audio_format[i].type;
+	}
+	return SPA_AUDIO_FORMAT_UNKNOWN;
+}
+
 #ifdef __cplusplus
 } /* extern "C" */
 #endif
