@@ -78,6 +78,8 @@ struct device {
 	uint32_t id;
 	uint32_t device_id;
 
+	char *factory_name;
+
 	struct rd_device *reserve;
 	struct spa_hook sync_listener;
 	int seq;
@@ -99,6 +101,7 @@ struct device {
 
 	unsigned int first:1;
 	unsigned int appeared:1;
+	unsigned int probed:1;
 	struct spa_list node_list;
 };
 
@@ -124,6 +127,8 @@ struct impl {
 #undef NAME
 #define NAME "alsa-monitor"
 
+static int probe_device(struct device *device);
+
 static struct node *alsa_find_node(struct device *device, uint32_t id)
 {
 	struct node *node;
@@ -159,8 +164,9 @@ static int node_acquire(void *data)
 	node->acquired = true;
 
 	if (device && device->n_acquired++ == 0 && device->reserve)
-		rd_device_acquire(device->reserve);
-	return 0;
+		return rd_device_acquire(device->reserve);
+	else
+		return 0;
 }
 
 static int node_release(void *data)
@@ -237,7 +243,7 @@ static struct node *alsa_create_node(struct device *device, uint32_t id,
 	}
 
 	priority = device->priority;
-	if (!strcmp(stream, "capture"))
+	if (node->direction == PW_DIRECTION_OUTPUT)
 		priority += 1000;
 	priority -= atol(dev) * 16;
 	priority -= atol(subdev);
@@ -271,9 +277,12 @@ static struct node *alsa_create_node(struct device *device, uint32_t id,
 	if (pw_properties_get(node->props, SPA_KEY_NODE_NAME) == NULL) {
 		const char *devname;
 		if ((devname = pw_properties_get(device->props, SPA_KEY_DEVICE_NAME)) == NULL)
-			devname = "unknown";
-		pw_properties_setf(node->props, SPA_KEY_NODE_NAME, "%s.%s.%s.%s",
-				devname, stream, dev, subdev);
+			devname = "unnamed-device";
+		if (strstr(devname, "alsa_card.") == devname)
+			devname += 10;
+		pw_properties_setf(node->props, SPA_KEY_NODE_NAME, "%s.%s.%s",
+				node->direction == PW_DIRECTION_OUTPUT ?
+				"alsa_input" : "alsa_output", devname, profile);
 	}
 	if (pw_properties_get(node->props, PW_KEY_NODE_DESCRIPTION) == NULL) {
 		const char *desc, *name = NULL;
@@ -576,14 +585,21 @@ static void reserve_acquired(void *data, struct rd_device *d)
 {
 	struct device *device = data;
 
-	pw_log_debug("%p: reserve acquired %d", device, device->n_acquired);
+	pw_log_info("%p: reserve acquired %d", device, device->n_acquired);
 
-	device->sdevice->locked = false;
+	if (!device->probed)
+		probe_device(device);
 
 	if (device->n_acquired == 0)
 		rd_device_release(device->reserve);
 }
 
+static void complete_release(struct device *device)
+{
+	if (device->reserve)
+		rd_device_complete_release(device->reserve, true);
+}
+
 static void sync_complete_done(void *data, int seq)
 {
 	struct device *device = data;
@@ -595,8 +611,7 @@ static void sync_complete_done(void *data, int seq)
 	spa_hook_remove(&device->sync_listener);
 	device->seq = 0;
 
-	if (device->reserve)
-		rd_device_complete_release(device->reserve, true);
+	complete_release(device);
 }
 
 static void sync_destroy(void *data)
@@ -616,7 +631,11 @@ static void reserve_release(void *data, struct rd_device *d, int forced)
 {
 	struct device *device = data;
 
-	pw_log_debug("%p: reserve release", device);
+	pw_log_info("%p: reserve release", device);
+	if (device->sdevice == NULL || device->sdevice->obj.proxy == NULL) {
+		complete_release(device);
+		return;
+	}
 
 	set_profile(device, 0);
 
@@ -632,7 +651,10 @@ static void reserve_busy(void *data, struct rd_device *d, const char *name, int3
 	struct device *device = data;
 	struct impl *impl = device->impl;
 
-	pw_log_debug("%p: reserve busy %s", device, name);
+	pw_log_info("%p: reserve busy %s", device, name);
+	if (device->sdevice == NULL)
+		return ;
+
 	device->sdevice->locked = true;
 
 	if (strcmp(name, "jack") == 0) {
@@ -647,7 +669,10 @@ static void reserve_available(void *data, struct rd_device *d, const char *name)
 	struct device *device = data;
 	struct impl *impl = device->impl;
 
-	pw_log_debug("%p: reserve available %s", device, name);
+	pw_log_info("%p: reserve available %s", device, name);
+	if (device->sdevice == NULL)
+		return ;
+
 	device->sdevice->locked = false;
 
 	remove_jack_timeout(impl);
@@ -688,6 +713,7 @@ static void device_free(void *data)
 	struct device *device = data;
 	pw_log_debug("device %p free", device);
 	spa_hook_remove(&device->listener);
+	free(device->factory_name);
 	pw_unload_spa_handle(device->handle);
 	pw_properties_free(device->props);
 	free(device);
@@ -719,15 +745,57 @@ static const struct sm_object_events device_events = {
 	.update = device_update,
 };
 
+static int probe_device(struct device *device)
+{
+	struct impl *impl = device->impl;
+	struct pw_context *context = impl->session->context;
+	struct spa_handle *handle;
+	void *iface;
+	int res;
+
+	handle = pw_context_load_spa_handle(context,
+			device->factory_name, &device->props->dict);
+	if (handle == NULL) {
+		res = -errno;
+		pw_log_error("can't make factory instance: %m");
+		goto exit;
+	}
+
+	if ((res = spa_handle_get_interface(handle, SPA_TYPE_INTERFACE_Device, &iface)) < 0) {
+		pw_log_error("can't get %s interface: %s", SPA_TYPE_INTERFACE_Device,
+				spa_strerror(res));
+		goto unload_handle;
+	}
+
+	device->handle = handle;
+	device->device = iface;
+
+	device->sdevice = sm_media_session_export_device(impl->session,
+			&device->props->dict, device->device);
+	if (device->sdevice == NULL) {
+		res = -errno;
+		goto unload_handle;
+	}
+	sm_object_add_listener(&device->sdevice->obj,
+			&device->listener,
+			&device_events, device);
+
+	device->probed = true;
+
+	return 0;
+
+unload_handle:
+	pw_unload_spa_handle(handle);
+exit:
+	return res;
+}
+
 static struct device *alsa_create_device(struct impl *impl, uint32_t id,
 		const struct spa_device_object_info *info)
 {
-	struct pw_context *context = impl->session->context;
 	struct device *device;
-	struct spa_handle *handle;
 	int res;
-	void *iface;
-	const char *card, *factory_name;
+	const char *str, *card, *factory_name, *name;
 
 	pw_log_debug("new device %u", id);
 
@@ -736,58 +804,46 @@ static struct device *alsa_create_device(struct impl *impl, uint32_t id,
 		return NULL;
 	}
 
+	device = calloc(1, sizeof(*device));
+	if (device == NULL) {
+		res = -errno;
+		goto exit;
+	}
+
 	if (impl->use_acp)
 		factory_name = SPA_NAME_API_ALSA_ACP_DEVICE;
 	else
 		factory_name = info->factory_name;
 
-	handle = pw_context_load_spa_handle(context,
-			factory_name, info->props);
-	if (handle == NULL) {
-		res = -errno;
-		pw_log_error("can't make factory instance: %m");
-		goto exit;
-	}
-
-	if ((res = spa_handle_get_interface(handle, info->type, &iface)) < 0) {
-		pw_log_error("can't get %s interface: %s", info->type, spa_strerror(res));
-		goto unload_handle;
-	}
-
-	device = calloc(1, sizeof(*device));
-	if (device == NULL) {
-		res = -errno;
-		goto unload_handle;
-	}
-
+	device->factory_name = strdup(factory_name);
 	device->impl = impl;
 	device->id = id;
-	device->handle = handle;
-	device->device = iface;
 	device->props = pw_properties_new_dict(info->props);
 	device->priority = 1000;
+	device->first = true;
+	spa_list_init(&device->node_list);
 	update_device_props(device);
+	device->pending_profile = 1;
+	spa_list_append(&impl->device_list, &device->link);
 
-	device->sdevice = sm_media_session_export_device(impl->session,
-			&device->props->dict, device->device);
-	if (device->sdevice == NULL) {
-		res = -errno;
-		goto clean_device;
+	if ((str = pw_properties_get(impl->session->props, "alsa.soft-mixer")) != NULL &&
+	    (strcmp(str, "*") == 0 ||
+	    ((name = pw_properties_get(device->props, "device.name")) != NULL &&
+	     strstr(str, name) != NULL))) {
+		pw_properties_set(device->props, "api.alsa.soft-mixer", "true");
 	}
 
-	sm_object_add_listener(&device->sdevice->obj,
-			&device->listener,
-			&device_events, device);
-
 	if (impl->conn &&
 	    (card = spa_dict_lookup(info->props, SPA_KEY_API_ALSA_CARD)) != NULL) {
 		const char *reserve;
 
+		device->priority -= atol(card) * 64;
+
 		pw_properties_setf(device->props, "api.dbus.ReserveDevice1", "Audio%s", card);
 		reserve = pw_properties_get(device->props, "api.dbus.ReserveDevice1");
 
 		device->reserve = rd_device_new(impl->conn, reserve,
-				"PipeWire", 10,
+				"PipeWire", -10,
 				&reserve_callbacks, device);
 
 		if (device->reserve == NULL) {
@@ -795,22 +851,14 @@ static struct device *alsa_create_device(struct impl *impl, uint32_t id,
 		} else {
 			rd_device_set_application_device_name(device->reserve,
 				spa_dict_lookup(info->props, SPA_KEY_API_ALSA_PATH));
-
-			rd_device_acquire(device->reserve);
 		}
-		device->priority -= atol(card) * 64;
 	}
-	device->pending_profile = 1;
-	device->first = true;
-	spa_list_init(&device->node_list);
-	spa_list_append(&impl->device_list, &device->link);
+	if (device->reserve != NULL)
+		rd_device_acquire(device->reserve);
+	else
+		probe_device(device);
 
 	return device;
-
-clean_device:
-	free(device);
-unload_handle:
-	pw_unload_spa_handle(handle);
 exit:
 	errno = -res;
 	return NULL;
