@@ -287,7 +287,9 @@ handle_node(struct impl *impl, struct sm_object *object)
 		else
 			return 0;
 
-		if (strcmp(media_class, "Sink") == 0)
+		if (strcmp(media_class, "Sink") == 0 ||
+		    strcmp(media_class, "Duplex") == 0 ||
+		    strcmp(media_class, "Source/Virtual") == 0)
 			direction = PW_DIRECTION_INPUT;
 		else if (strcmp(media_class, "Source") == 0)
 			direction = PW_DIRECTION_OUTPUT;
