@@ -35,6 +35,7 @@
 #include <spa/utils/result.h>
 
 #include <pipewire/impl.h>
+#include <pipewire/private.h>
 
 #define NAME "access"
 
@@ -188,6 +189,7 @@ context_check_access(void *data, struct pw_impl_client *client)
 	    (access = pw_properties_get(impl->properties, "access.force")) != NULL)
 		goto wait_permissions;
 
+#if defined(__linux__)
 	res = check_flatpak(client, pid);
 	if (res != 0) {
 		if (res < 0) {
@@ -199,23 +201,25 @@ context_check_access(void *data, struct pw_impl_client *client)
 			}
 		}
 		else if (res > 0) {
-			pw_log_debug(NAME" %p: sandboxed client %p added", impl, client);
+			pw_log_debug(NAME" %p: flatpak client %p added", impl, client);
 		}
 		access = "flatpak";
 		goto wait_permissions;
 	}
+#endif
 	access = "unrestricted";
 
 granted:
 	pw_log_info(NAME" %p: client %p '%s' access granted", impl, client, access);
 	items[0] = SPA_DICT_ITEM_INIT(PW_KEY_ACCESS, access);
 	pw_impl_client_update_properties(client, &SPA_DICT_INIT(items, 1));
+
 	permissions[0] = PW_PERMISSION_INIT(PW_ID_ANY, PW_PERM_ALL);
 	pw_impl_client_update_permissions(client, 1, permissions);
 	return;
 
 wait_permissions:
-	pw_log_debug(NAME " %p: client %p wait for '%s' permissions",
+	pw_log_info(NAME " %p: client %p wait for '%s' permissions",
 			impl, client, access);
 	items[0] = SPA_DICT_ITEM_INIT(PW_KEY_ACCESS, access);
 	pw_impl_client_update_properties(client, &SPA_DICT_INIT(items, 1));
