@@ -73,6 +73,9 @@ static void info_changed(struct pw_impl_link *link)
 {
 	struct pw_resource *resource;
 
+	if (link->info.change_mask == 0)
+		return;
+
 	pw_impl_link_emit_info_changed(link, &link->info);
 
 	if (link->global)
@@ -109,7 +112,10 @@ static void link_update_state(struct pw_impl_link *link, enum pw_link_state stat
 	pw_impl_link_emit_state_changed(link, old, state, error);
 
 	link->info.change_mask |= PW_LINK_CHANGE_MASK_STATE;
-	info_changed(link);
+	if (state == PW_LINK_STATE_ERROR ||
+	    state == PW_LINK_STATE_PAUSED ||
+	    state == PW_LINK_STATE_ACTIVE)
+		info_changed(link);
 
 	if (state == PW_LINK_STATE_ERROR && link->global) {
 		struct pw_resource *resource;
@@ -117,7 +123,7 @@ static void link_update_state(struct pw_impl_link *link, enum pw_link_state stat
 			pw_resource_error(resource, res, error);
 	}
 
-	if (old != PW_LINK_STATE_PAUSED && state == PW_LINK_STATE_PAUSED) {
+	if (old < PW_LINK_STATE_PAUSED && state == PW_LINK_STATE_PAUSED) {
 		link->prepared = true;
 		link->preparing = false;
 		pw_context_recalc_graph(link->context, "link prepared");
@@ -223,12 +229,12 @@ static int do_negotiate(struct pw_impl_link *this)
 		case -EIO:
 			current = NULL;
 			res = 0;
-			/* fallthrough */
+			SPA_FALLTHROUGH
 		case 1:
 			break;
 		case 0:
 			res = -EBADF;
-			/* fallthrough */
+			SPA_FALLTHROUGH
 		default:
 			error = spa_aprintf("error get output format: %s", spa_strerror(res));
 			goto error;
@@ -257,12 +263,12 @@ static int do_negotiate(struct pw_impl_link *this)
 		case -EIO:
 			current = NULL;
 			res = 0;
-			/* fallthrough */
+			SPA_FALLTHROUGH
 		case 1:
 			break;
 		case 0:
 			res = -EBADF;
-			/* fallthrough */
+			SPA_FALLTHROUGH
 		default:
 			error = spa_aprintf("error get input format: %s", spa_strerror(res));
 			goto error;
@@ -328,10 +334,9 @@ static int do_negotiate(struct pw_impl_link *this)
 	free(this->info.format);
 	this->info.format = format;
 
-	if (changed) {
+	if (changed)
 		this->info.change_mask |= PW_LINK_CHANGE_MASK_FORMAT;
-		info_changed(this);
-	}
+
 	pw_log_debug(NAME" %p: result %d", this, res);
 	return res;
 
@@ -548,6 +553,7 @@ int pw_impl_link_activate(struct pw_impl_link *this)
 
 	impl->activated = true;
 	pw_log_info("(%s) activated", this->name);
+	link_update_state(this, PW_LINK_STATE_ACTIVE, 0, NULL);
 
 	return 0;
 }
@@ -561,7 +567,7 @@ static void check_states(void *obj, void *user_data, int res, uint32_t id)
 	if (this->info.state == PW_LINK_STATE_ERROR)
 		return;
 
-	if (this->info.state == PW_LINK_STATE_PAUSED)
+	if (this->info.state >= PW_LINK_STATE_PAUSED)
 		return;
 
 	output = this->output;
@@ -716,6 +722,7 @@ int pw_impl_link_deactivate(struct pw_impl_link *this)
 	impl->io_set = false;
 	impl->activated = false;
 	pw_log_info("(%s) deactivated", this->name);
+	link_update_state(this, PW_LINK_STATE_PAUSED, 0, NULL);
 
 	return 0;
 }
@@ -735,7 +742,7 @@ global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
 	pw_log_debug(NAME" %p: bound to %d", this, resource->id);
 	pw_global_add_resource(global, resource);
 
-	this->info.change_mask = ~0;
+	this->info.change_mask = PW_LINK_CHANGE_MASK_ALL;
 	pw_link_resource_info(resource, &this->info);
 	this->info.change_mask = 0;
 
@@ -1194,6 +1201,8 @@ int pw_impl_link_register(struct pw_impl_link *link,
 		PW_KEY_CLIENT_ID,
 		PW_KEY_LINK_OUTPUT_PORT,
 		PW_KEY_LINK_INPUT_PORT,
+		PW_KEY_LINK_OUTPUT_NODE,
+		PW_KEY_LINK_INPUT_NODE,
 		NULL
 	};
 
@@ -1275,6 +1284,8 @@ void pw_impl_link_destroy(struct pw_impl_link *link)
 	pw_log_debug(NAME" %p: free", impl);
 	pw_impl_link_emit_free(link);
 
+	spa_hook_list_clean(&link->listener_list);
+
 	pw_work_queue_destroy(impl->work);
 
 	pw_properties_free(link->properties);
